//This file is meant to be left as is, except for the device settings (a few lines below this line)

#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <ESP32Servo.h>  // older version <ESP32_Servo.h>, newer version same library <ESP32Servo.h>
#include <WString.h>
#include "DNSServer.h"

#include "car.h"
#include "webPage.h"
#include "functions.h"

namespace car{
  /////////////// Device Settings - you can change //////////////////
  
  String ssid_Host     = "DISK";
  bool concatinate_mac_address_to_ssid_Host = true;
  const char* password_Host = "passwordsecret";
  bool passwordProtect = false;

// We do not want to connect to existing network, now each robot is an unique access point
//    --> if you enable this, you can controll the robot and use internet at the same time; you may need to guess the IP nbr of your robot
  const char* ssid_Client     = "mijnthuiswifi";
  const char* password_Client = "pass";
  bool connectToWiFi = false;  //The program becomes sluggish when this setting is set to true and there is no WiFi network available
  //If connectToWiFi is set to true, the captive portal dissapears

  // the next variable is used in an interrupt, so it should be volatile
  volatile int timeout = 5000; //timeout in milliseconds, so if the robot doesn't get a command in this timeframe, it should stop less then 100 ms is not possible
  // the timeout stops the robot
  
  //                  L   L   R   R
  int motorPins[4] = {18, 5, 13, 15};


  //////////////// Defines - do not change //////////////
  bool logData = true;
  
  int throttle = 0;
  int rot     = 0;

  // the next variable is used in an interrupt, so it should be volatile
  volatile long time_since_last_websockets_data = 0;

  //declare timer for interrupting
  hw_timer_t* timer = NULL;  
  
  AsyncWebServer server(80);
  AsyncWebSocket ws("/ws");

  TaskHandle_t carTask;
  
  //dns for captive login
  DNSServer dnsServer;                      //This creates a DNS server, required for the captive portal


  ///////////////Functions/////////////
  void initCar(){        
    //init the motors by attaching the pins
    initMotor();

    //setup serial
    Serial.begin(115200);
    Serial.println();

    //setup timeout interrupt
    timer = timerBegin(0, 80, true);
    timerAttachInterrupt(timer, &checkTimeout, true);
    timerAlarmWrite(timer, 100000, true); //run function every 100 ms
    timerAlarmEnable(timer);

    //concatinate the mac address to the ssid if settings are set for it
    if(concatinate_mac_address_to_ssid_Host){
      ssid_Host = String(ssid_Host) + String("-") + String((uint32_t)ESP.getEfuseMac(), HEX);
    }

    //start wifi network host
    if(passwordProtect){
      WiFi.softAP(ssid_Host.c_str(), password_Host);
    }else{
      WiFi.softAP(ssid_Host.c_str());
    }

    //connect to wifi network as client
    if(connectToWiFi) WiFi.begin(ssid_Client, password_Client);

    //print the IP´s of the device
    IPAddress HostIP = WiFi.softAPIP();
    Serial.print("IP on WiFi hosted network " + String(ssid_Host) + ": '");
    Serial.print(HostIP);
    Serial.println("'");

    if(connectToWiFi){
      IPAddress ClientIP = WiFi.localIP();
      Serial.print("IP on WiFi connected network " + String(ssid_Client) + ": '");
      Serial.print(ClientIP);
      Serial.println("'");
    }else{
      Serial.println("We are not connected to a WiFi network");
    }

    //configure the webserver (websockets and webpages)
    ws.onEvent(onWebsocket);
    server.addHandler(&ws);
    
    server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
      request->send(200, "text/html", webPage::pageC);
    });

    server.on("/generate_204", HTTP_GET, [](AsyncWebServerRequest *request){
      // Don't redirect to the main page because andriod will use the built in browser which is not compatible with the webpage
      request->send(200, "text/html", "<h1>Please connect to http://www.junioriot.nl<\h1>");
      Serial.println("requested /generate_204");
    });

    server.onNotFound([](AsyncWebServerRequest *request){
      request->redirect("/");
    });

    //Setup dns for captive login page
    if(connectToWiFi == false){
      xTaskCreate(
                      dnsTask,          /* Task function. */
                      "dnsTask",        /* String with name of task. */
                      10000,            /* Stack size in bytes. */
                      NULL,             /* Parameter passed as input of the task */
                      1,                /* Priority of the task. */
                      NULL);            /* Task handle. */
    }

    //start the webserver
    server.begin();
    
    //run setup function from functions.cpp
    setupFunction();

    // run loop function from functions.cpp
    while(true){
      loopFunction();
    }
  }

  //websocket handle function
  void onWebsocket(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len){
    AwsFrameInfo * info = (AwsFrameInfo*)arg;

    if(type == WS_EVT_CONNECT){
      Serial.println("Client '" + String(client->id()) + "' has connected");
    } else if(type == WS_EVT_PONG){
      Serial.println("Client '" + String(client->id()) + "' has answered ping request");
    } else if(type == WS_EVT_DISCONNECT){
      Serial.println("Client '" + String(client->id()) + "' has been disconnected");
      movement(0, 0); //Stop the robot when client is disconnected
    } else if(type == WS_EVT_ERROR){
      Serial.println("Client '" + String(client->id()) + "': an error occured");
      movement(0, 0); //Stop the robot when client is disconnected
    } else if(type == WS_EVT_DATA){
      if(info->opcode == WS_TEXT){
        //text was received, this is done by the javascript code in our webpage

        //reset the timeout timer
        time_since_last_websockets_data = millis();
        
        //find the command sent by the js code
        int key = (char)data[0];
        
        //build the value string out of received chars
        String value;
        for(size_t i=1; i < info->len; i++){
          value += (char)data[i];
        }

        //run the code which handles the commands sent by the js code
        switch(key){
          case 't':
            throttle = value.toInt();
            break;
          case 'r':
            rot = value.toInt();
            break;
          case 'b':
            if(value.toInt() == 1){
              button1();
            }else if(value.toInt() == 2){
              button2();
            }else if(value.toInt() == 3){
              button3();
            }else if(value.toInt() == 4){
              button4();
            }
            break;
        }
        movement(throttle, rot);
      }else{
        //binary message was received, do nothing, our js code shouldn´t send binary messages
      }
    }
  }

  //check timeout, as an interrupt
  void IRAM_ATTR checkTimeout(){                                          
    if(millis() - time_since_last_websockets_data > timeout){
      movement(0, 0); //Stop the robot
    }
  }

  //task for dns server handle
  void dnsTask( void * parameter ){
    dnsServer.start(53, "*", WiFi.softAPIP());  //This starts the DNS server.  The "*" sends any request for port 53 straight to the IP address of the device
    while(true){
      dnsServer.processNextRequest();         //Without this, the connected device will simply timeout trying to reach the internet
                                            //or it might fall back to mobile data if it has i
      delay(100); //free up the core
    }
  }

  //init the motors by attaching the correct pins to the pwm device (internal in esp)
  void initMotor(){
    for(int i = 1; i <= 4; i++){
      ledcSetup(i+4, 5000, 8);  //use channel 5, 6, 7, 8
      ledcAttachPin(motorPins[i - 1], i+4);
    }
  }

  //convert integer into motor compatable value (by setting the min and max values)
  int toMotorSpeed(int x){
    if(x < -255) x = -255;
    if(x > 255) x = 255;
    return x;
  }

  //return the Throttle defined by webpage
  int getThrottle(){
    return throttle;
  }

  //return the rotation defined by webpage
  int getRotation(){
    return rot;
  }

  //write motor Throttle value to motor
  void writeMotorL(int motorThrottle){
    //make sure Throttle value is compatable with motor
    motorThrottle = toMotorSpeed(motorThrottle);

    //find movement rotation and write the speed to the correct motor pin
    if(motorThrottle > 0){
      ledcWrite(1+4, motorThrottle);
      ledcWrite(2+4, 0);
    }else{
      ledcWrite(1+4, 0);
      ledcWrite(2+4, -motorThrottle); // motorThrottle is negative, -motorThrottle is positive
    }
  }

  //write motor Throttle value to motor
  void writeMotorR(int motorThrottle){
    //make sure Throttle value is compatable with motor
    motorThrottle = toMotorSpeed(motorThrottle);

    //find movement rotation and write the speed to the correct motor pin
    if(motorThrottle > 0){
      ledcWrite(3+4, motorThrottle);
      ledcWrite(4+4, 0);
    }else{
      ledcWrite(3+4, 0);
      ledcWrite(4+4, -motorThrottle);    
    }
  }

  //write both motors in the same function call
  void writeMotors(int motorL, int motorR){
    writeMotorL(motorL);
    writeMotorR(motorR);
  }
}
